/*
 * Copyright (c) 2011, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

/*
 * event.c -- event system
 *
 * The enqueued events are stored in a ring buffer of 2^n entries (to
 * implement "wrap around" using a bit mask for performance reasons).
 * The callback subscriptions are stored in an array, using a bit mask to
 * encode the events a subscriber is interested in. The size of this bit
 * mask may need to be increased when more event IDs are added.
 * The callback function makes the client unique; subsequent registrations
 * with the same function don't create new entries, but modify the event
 * bit mask. If during an unregister operation the last bit in this mask
 * is cleared, the callback entry itself is removed either (by resetting
 * the function pointer as well).
 *
 */

#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>

#include "config.h"
#include "util.h"
#include "event.h"

/* Validation of configuration settings */
#if !(defined EVT_QUEUE_LEN && (EVT_QUEUE_LEN > 1) && IS_POWER_OF_TWO(EVT_QUEUE_LEN))
#error "EVT_QUEUE_LEN must be defined to 2^n with n >= 1"
#endif 

/* Event queue, implemented as ring buffer.
 * Access must be synchronized (shared between ISR and normal context).
 */
static volatile struct {
	uint8_t   start;  // index of first valid entry
	uint8_t   count;  // number of valid entries
	evt_id_t  events[EVT_QUEUE_LEN];
} evt_queue;

/* Type of bit mask for event subscriptions */
typedef uint8_t evt_mask_t;
/* Entry of an event subscription */
typedef struct {
	evt_cb_t    func;    // callback function
	evt_mask_t  idmask;  // event bit mask
} evt_subscriber_t;

/* Compile time assertion to assure that all events fit into bit mask */
CT_ASSERT(EVT_NUM_EVENTS <= (sizeof(evt_mask_t) * 8));

/* Event subscription list. Valid entries are compacted at the beginning
 * of the array. Empty entries have a callback function of NULL.
 */
static evt_subscriber_t evt_subscr_list[EVT_MAX_SUBSCR];

/* Initialization function */
void evt_init(void)
{
	/* Reset event queue */
	evt_queue.start = 0;
	evt_queue.count = 0;
	
	/* Clear out all subscriber entries */
	for (uint8_t idx = 0; idx < EVT_MAX_SUBSCR; idx++) {
		evt_subscr_list[idx].func = NULL;
		evt_subscr_list[idx].idmask = 0;
	}
}

/* Enqueue an event */
error_t evt_enqueue(evt_id_t id)
{
	error_t result = EAGAIN;
	
	/* Events may be enqueued from different contexts */
	uint8_t sreg;
	irq_lock(&sreg);
	
	/* Assert that there is enough place in the queue */
	RT_ASSERT(evt_queue.count < EVT_QUEUE_LEN);
	if (evt_queue.count < EVT_QUEUE_LEN) {
		/* Insert new entry at the end of the queue,
		 * index is wrapped if needed.
		 */
		uint8_t end = (evt_queue.start + evt_queue.count) & (EVT_QUEUE_LEN - 1);
		evt_queue.events[end] = id;
		evt_queue.count++;
		
		result = ENOERR;
	}
	
	irq_unlock(sreg);
	return result;
}

/* Dequeue an event for dispatching (used internally only) */
static error_t evt_dequeue(evt_id_t *id)
{
	error_t result = EAGAIN;
	
	/* Must be synchronized with enqueue operations */
	uint8_t sreg;
	irq_lock(&sreg);
	
	if (evt_queue.count > 0) {
		/* Get event from beginning of queue and advance index.
		 * Wrap around if needed.
		 */
		*id = evt_queue.events[evt_queue.start];
		evt_queue.start = (evt_queue.start + 1) & (EVT_QUEUE_LEN - 1);
		evt_queue.count--;
		
		result = ENOERR;
	}
	
	irq_unlock(sreg);
	return result;
}

/* Register subscriber for event */
error_t evt_register(evt_cb_t func, evt_id_t id)
{
	error_t result = ENOMEM;
	
	evt_subscriber_t *subscr = &evt_subscr_list[0];
	evt_subscriber_t *empty = NULL;
	evt_subscriber_t *match = NULL;
	
	/* Find slot to insert registration */
	while (subscr != &evt_subscr_list[EVT_MAX_SUBSCR]) {
		if (subscr->func == func) {
			/* Direct match, do not search further */
			match = subscr;
			break;
		}
		if (!empty && (subscr->func == NULL)) {
			/* Remember first empty slot */
			empty = subscr;
		}
		subscr++;
	}
	
	/* Prefer direct match, otherwise empty slot */
	evt_subscriber_t *slot = match ? match : empty;
	
	/* If valid slot was found, insert data */
	if (slot) {
		slot->func = func;
		slot->idmask |= (1 << id);
		result = ENOERR;
	}
	
	/* Assure in debug version that registration was successful */
	RT_ASSERT(result == ENOERR);
	return result;
}

/* Unregisters for an event */
error_t evt_unregister(evt_cb_t func, evt_id_t id)
{
	error_t result = EINVAL;
	
	evt_subscriber_t *subscr = &evt_subscr_list[0];
	bool compact = false;
	const uint8_t idmask = (1 << id);
	
	while (subscr != &evt_subscr_list[EVT_MAX_SUBSCR]) {
		if (compact) {
			/* Compact list by pushing entries together */
			*(subscr - 1) = *subscr;
			subscr->func = NULL;
			subscr->idmask = 0;
		} else if ((subscr->func == func) && (subscr->idmask & idmask)) {
			/* Entry found, remove ID from mask */
			subscr->idmask &= ~idmask;
			/* If mask is empty now, remove callback as well */
			if (subscr->idmask == 0) {
				subscr->func = NULL;
				compact = true;
			}
			result = ENOERR;
		}
		subscr++;
	}

	/* Assure in debug version that unregistration was successful */
	RT_ASSERT(result == ENOERR);	
	return result;
}

/* Event dispatching in normal context */
void evt_dispatch(void)
{
	evt_id_t event = 0;
	
	/* Dequeue events until queue is empty */
	while (evt_dequeue(&event) == ENOERR) {
		const evt_subscriber_t *subscr = &evt_subscr_list[0];
		/* Iterate over all subscribers */
		while (subscr->func && (subscr != &evt_subscr_list[EVT_MAX_SUBSCR])) {
			/* Call interested clients only */
			if (subscr->idmask & (1 << event))
				subscr->func(event);
			subscr++;
		}
	}
}
