/*
 * Copyright (c) 2011, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

/*
 * event.h -- event system
 *
 * The event systems allows clients to register for specific events. The
 * callback function provided during registration is called for each event
 * it was registered for. A callback function can be registered for multiple
 * events. The ID of the event is passed as parameter to the callback
 * function to distinguish events in this case.
 *
 * The current implementation uses non-blocking enqueue operations only. An
 * error of EAGAIN is returned if an enqueue operations fails because the
 * queue is full.
 *
 * Registration and unregistration is allowed from normal context only.
 *
 * Enqueuing an event is allowed from all contexts (e.g. from interrupt
 * service routines).
 *
 * Event dispatching is allowed from normal context only.
 *
 */

#ifndef EVENT_H
#define EVENT_H

#include "types.h"

/* Enumeration of all event IDs */
typedef enum {
	EVT_UART_RXCH,    // received characters are available
	EVT_UART_RXLN,    // a received line is completed
	EVT_TMR_1MS,      // the 1ms timer has elapsed
	EVT_PULSE_BEAT,   // a heart beat has been detected
	EVT_NUM_EVENTS    // number of events (used internally)
} evt_id_t;

/* Callback function used for registration
 *
 * Parameters:
 *   id    event ID that caused callback
 */
typedef void (*evt_cb_t)(evt_id_t id);

/* Initialization function */
void evt_init(void);

/* Enqueues an event. May be called from all contexts.
 *
 * Parameters:
 *   id        event ID to enqueue
 *
 * Return value:
 *   ENOERR    action was successful
 *   EAGAIN    action would block (event not enqueued)
 */
error_t evt_enqueue(evt_id_t id);

/* Registers for an event. Must be called from normal context only.
 *
 * Parameters:
 *   func    callback function to register
 *   id      event ID to register for
 *
 * Return value:
 *   ENOERR  action was successful
 *   ENOMEM  no memory left to perform registration
 */
error_t evt_register(evt_cb_t func, evt_id_t id);

/* Unregisters a previous registration.
 * Must be called from normal context only.
 *
 * Parameters:
 *   func    callback function to unregister
 *   id      event ID to unregister for
 *
 * Return value:
 *   ENOERR  action was successful
 *   EINVAL  invalid arguments, e.g. callback was not registered
 */
error_t evt_unregister(evt_cb_t func, evt_id_t id);

/* Dispatches all enqueued events. Must be called from normal context only.
 * Note that dispatching stops only when queue is empty, e.g. events enqueued
 * while dispatching will be dispatched, too.
 */
void evt_dispatch(void);

#endif
