/*
 * Copyright (c) 2011, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

/*
 * pulse.c -- pulse detection
 *
 * The external hardware pulls a port pin to LOW on each detected heart beat.
 * Depending on the configuration this either generates an external interrupt
 * or is monitored by sampling the pin state every millisecond. When using
 * interrupts, each one is considered to be an event. In polling mode a simple
 * debounce algorithm is applied. If the pin state is considered as stable, an
 * event is generated.
 *
 * The interval between the events is measured and the corresponding heart
 * rate in beats per minute (bpm) is calculated. Since these values are
 * varying an average is calculated using a sliding window algorithm (last N
 * values are stored and averaged using the arithmetic mean).
 *
 * The values are considered as invalid if the interval rises above a
 * threshold defined by PULSE_INVALIDATE_TIMEOUT.
 */

#include <avr/io.h>
#include <stdbool.h>
#include <stdint.h>

#include "config.h"
#include "event.h"
#include "pulse.h"

#ifdef PULSE_USE_INT
#include <avr/interrupt.h>
#include "util.h"
/* Used to protect variables for access from
 * interrupt context and normal context.
 */
#define VOLATILE        volatile
#define SREG_BUFFER(x)  uint8_t x
#define IRQ_LOCK(x)     irq_lock(x)
#define IRQ_UNLOCK(x)   irq_unlock(x)
#else
/* No interrupts, no access protection needed */
#define VOLATILE
#define SREG_BUFFER(x)
#define IRQ_LOCK(x)
#define IRQ_UNLOCK(x)
/* Debounce counter */
static uint8_t pulse_debounce;
#endif

/* Timeout after which the results are invalidated */
#define PULSE_INVALIDATE_TIMEOUT  5000

/* Tick counter for measuring the heart beat intervals */
static VOLATILE uint16_t pulse_ticks;

/* Data structure to store heart rate values. The value at index last_idx
 * is always the most current value. Invalid entries are set to 0.
 */
static VOLATILE struct
{
	uint8_t  last_idx;  // last index (= most current value)
	uint8_t  values[PULSE_AVG_COUNT];
} pulse_data;

/* Called on every detected event on input pin */
static void pulse_detected(void)
{
	/* Preset with invalid marker */
	uint8_t bpm = 0;
	
	/* Consider result only if tick count is below invalidate threshold.
	 * Tick count must be greater than 0 to avoid division by zero.
	 */
	if ((pulse_ticks < PULSE_INVALIDATE_TIMEOUT) && (pulse_ticks > 0))
		bpm = 60000 / pulse_ticks;
	
	/* Store new result in ring buffer, wrap around if needed */
	pulse_data.last_idx++;
	if (pulse_data.last_idx >= PULSE_AVG_COUNT)
		pulse_data.last_idx = 0;
	pulse_data.values[pulse_data.last_idx] = bpm;
	
	/* Reset tick count for next cycle */
	pulse_ticks = 0;
	/* Enqueue heart beat event */
	evt_enqueue(EVT_PULSE_BEAT);
}

#ifdef PULSE_USE_INT
/* Interrupt handler for falling edges on INT0 pin */
ISR(INT0_vect)
{
	pulse_detected();
}
#endif

/* Callback function for 1ms timer events */
static void pulse_1ms_tick(evt_id_t id)
{
	SREG_BUFFER(sreg);
	
	/* Interrupt lock because of shared variable hrs_ticks */
	IRQ_LOCK(&sreg);

	/* Increment tick count until invalidate threshold is reached */
	if (pulse_ticks < PULSE_INVALIDATE_TIMEOUT)
		pulse_ticks++;

#ifndef PULSE_USE_INT
	/* Sample input pin */
	bool state = (PULSE_SENSE_IREG & (1 << PULSE_SENSE_PIN)) != 0;

	/* Debounce algorithm */
	if (state) {
		if (pulse_debounce > 0) {
			pulse_debounce--;
			if (pulse_debounce == 0)
				pulse_detected();
		}
	} else {
		pulse_debounce = PULSE_SENSE_COUNT;
	}
#endif

	IRQ_UNLOCK(sreg);
}

/* Initialization function */
void pulse_init(void)
{
	/* Reset data structure.
	 * Invalidating the first entry is sufficient because it will be
	 * soonest overwritten after one wrap-around cycle. One invalid value
	 * is also enough to have the average value invalidated.
	 */
	pulse_data.last_idx = 0;
	pulse_data.values[0] = 0;

#ifdef PULSE_USE_INT
	/* Interrupts on falling edge on INT0 pin */
	MCUCR &= ~(1 << ISC00);
	MCUCR |= (1 << ISC01);
	/* Enable INT0 */
	GICR |= (1 << INT0);
#else
	/* Set debounce counter to 0. This requires the port pin to be
	 * HIGH before a new debounce operation is started and thus
	 * prevents a false trigger when the device is booted up with
	 * a static LOW level on the input pin.
	 */
	pulse_debounce = 0;
	/* Configure port pin as input with pull-up enabled */
	PULSE_SENSE_DREG &= ~(1 << PULSE_SENSE_PIN);
	PULSE_SENSE_OREG |= (1 << PULSE_SENSE_PIN);
#endif
	
	/* Set tick counter to invalid until first interrupt occurs */
	pulse_ticks = PULSE_INVALIDATE_TIMEOUT;
	/* Register event callback */
	evt_register(pulse_1ms_tick, EVT_TMR_1MS);
}

/* Get current heart rate value */
bool pulse_get_curr(uint8_t *curr)
{
	bool valid;
	SREG_BUFFER(sreg);
	
	/* Interrupt lock because of shared variable hrs_data */
	IRQ_LOCK(&sreg);
	
	/* Get most current value */
	*curr = pulse_data.values[pulse_data.last_idx];
	/* Result is valid if value is not the invalid marker */
	valid = (*curr != 0);
	
	IRQ_UNLOCK(sreg);
	return valid;
}

/* Get average heart rate value */
bool pulse_get_avg(uint8_t *avg)
{
	bool valid;
	SREG_BUFFER(sreg);
	
	/* Interrupt lock because of shared variable hrs_data */
	IRQ_LOCK(&sreg);
	
	uint8_t curr;
	/* Variable for sum must be capable to fit at least HRS_AVG_COUNT
	 * times the maximum heart rate of 255, which is true for uint16_t.
	 */
	uint16_t sum = 0;
	/* Result is considered as valid unless the invalid marker is found */
	valid = true;
	/* Loop over all results, order is irrelevant */
	for (uint8_t idx = 0; idx < PULSE_AVG_COUNT; idx++) {
		/* Check if current value is valid, otherwise reset flag */
		curr = pulse_data.values[idx];
		if (curr == 0)
			valid = false;
		/* Add current value to sum */
		sum += curr;
	}
	/* Calculate the arithmetic mean */
	*avg = sum / PULSE_AVG_COUNT;
	
	IRQ_UNLOCK(sreg);
	return valid;	
}
