/*
 * Copyright (c) 2011, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

/*
 * timer.c -- Timer event generation
 *
 * The hardware timer 0 is used to generate timer events. Currently the 1ms
 * event is generated only. Applications can implement software timers on
 * base of this event. Support for user-defined timers could be added if there
 * is sufficient need for such a feature.
 *
 */

#include <avr/io.h>
#include <avr/interrupt.h>

#include "config.h"
#include "event.h"
#include "timer.h"

/* Best performance:
 * F_CPU     Presc.  HW div.  SW div.  Freq.
 * 16000000  64      250      1        1000
 *  8000000  64      125      1        1000
 *  4000000  8       250      2        1000
 *  2000000  8       250      1        1000
 *  1000000  8       125      1        1000
 *
 * Most flexible configuration:
 * F_CPU     Presc.  HW div.  SW div.  Freq.
 * 16000000  8       125      16       1000
 *  8000000  8       125      8        1000
 *  4000000  8       125      4        1000
 *  2000000  8       125      2        1000
 *  1000000  8       125      1        1000
 */

#define TMR_PRESC  (1 << CS01)
#define TMR_HWDIV  125
#define TMR_SWDIV  (F_CPU / 1000000UL)

/* Used from ISR context only! */
static uint8_t tmr_swdiv;

/* Interrupt handler for overflow */
ISR(TIMER0_OVF_vect)
{
	/* Reload counter register */
	TCNT0 = 256 - TMR_HWDIV;
	
	/* Software divisor */
	if (tmr_swdiv > 0)
		tmr_swdiv--;
	if (tmr_swdiv == 0) {
		evt_enqueue(EVT_TMR_1MS);
		tmr_swdiv = TMR_SWDIV;
	}
}

/* Initialization function */
void tmr_init(void)
{
	/* Set prescaler */
	TCCR0 = TMR_PRESC;
	/* Enable overflow interrupt */
	TIMSK |= (1 << TOIE0);
}
