/*
 * Copyright (c) 2011, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

/*
 * uart.c -- UART services
 *
 * The UART is used for sending and receiving data. Baud rate and mode
 * (text/binary) are configured in config.h. The frame layout is always
 * 8 data bits, no parity, 1 stop bit (aka. 8N1).
 *
 * There are separate buffers for outgoing and incoming data (size must be 2^n
 * with n >= 1, configured in config.h), implemented as ring buffer. When
 * sending the Data Register Empty interrupt is used to move the next byte
 * from the output buffer into the Data Register. Incoming data is stored in
 * the RX Complete interrupt service routine. If no buffer space is available,
 * the new data is silently discarded (run-time assertion in debug mode).
 *
 * High-level access functions compatible to ANSI's fputc/fgetc/fputs/fgets
 * are provided. Sending functions block if the output buffer is full,
 * receiving functions treat an empty input buffer as EOF. If a new line as
 * required by fgets is complete, the EVT_UART_RXLN event is sent.
 * Additionally the EVT_UART_RXCH event is sent if a new character data has
 * arrived. EVT_UART_RXCH is sent only once until the uart_rx_byte function is
 * called (e.g. indirectly by using fgetc/fgets).
 *
 * In binary mode (UART_MODE_BINARY) the data is passed as it is received.
 *
 * In text mode (UART_MODE_TEXT) filter procedures are used for CR/LF fix up.
 * For outgoing text data LF is expanded into CRLF (CR is passed unchanged).
 * Incoming text data is expected to contain at least CR as line termination,
 * LF is discarded. The high level interface thus always terminates lines with
 * a single LF ('\n'), as expected for the "f-functions" in text mode.
 *
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <stdbool.h>

#include "config.h"
#include "util.h"
#include "event.h"
#include "uart.h"

/* Validation of configuration settings */
#if !(defined UART_TX_BUFSIZE && (UART_TX_BUFSIZE > 1) && IS_POWER_OF_TWO(UART_TX_BUFSIZE))
#error "UART_TX_BUFSIZE must be defined to 2^n with n >= 1"
#endif 
#if !(defined UART_RX_BUFSIZE && (UART_RX_BUFSIZE > 1) && IS_POWER_OF_TWO(UART_RX_BUFSIZE))
#error "UART_RX_BUFSIZE must be defined to or 2^n with n >= 1"
#endif 
#if (!defined UART_MODE_TEXT) && (!defined UART_MODE_BINARY)
#error "Either UART_MODE_TEXT or UART_MODE_BINARY must be defined"
#endif
#if (defined UART_MODE_TEXT) && (defined UART_MODE_BINARY)
#error "UART_MODE_TEXT and UART_MODE_BINARY must not be defined both"
#endif

/* Buffer for outgoing data. Used by ISR and normal context! */
static volatile struct {
	uint8_t  start;  // index of first valid data item
	uint8_t  count;  // number of valid data items
	uint8_t  data[UART_TX_BUFSIZE];
} uart_txbuf;

/* Buffer for incoming data. Used by ISR and normal context! */
static volatile struct
{
	uint8_t  start;  // index of first valid data item
	uint8_t  count;  // number of valid data items
	uint8_t  data[UART_RX_BUFSIZE];
} uart_rxbuf;

/* Flag to remember if EVT_UART_RXCH has been sent.
 * Used by ISR and normal context!
 */
static volatile bool notified;

/* Filter procedure for input */
inline bool uart_filter_input(uint8_t *byte)
{
	bool pass = true;
	
#ifdef UART_MODE_TEXT
	switch (*byte) {
	case '\n':
		/* LF is discarded */
		pass = false;
		break;
	case '\r':
		/* CR is passed as LF */
		*byte = '\n';
		break;
	}
#endif
	
#ifdef UART_MODE_BINARY
	/* Nothing to do */
#endif
	
	return pass;
}

/* Data Register Empty interrupt vector */
ISR(USART_UDRE_vect)
{
	if (uart_txbuf.count > 0) {
		/* send next byte from buffer */
		UDR = uart_txbuf.data[uart_txbuf.start];
		uart_txbuf.start = (uart_txbuf.start + 1) & (UART_TX_BUFSIZE - 1);
		uart_txbuf.count--;
	} else {
		/* disable UDRE interrupt */
		UCSRB &= ~(1 << UDRIE);
	}
}

/* Receive Complete interrupt vector */
ISR(USART_RXC_vect)
{
	uint8_t byte = UDR;
	
	if (uart_rxbuf.count < UART_RX_BUFSIZE) {
		if (uart_filter_input(&byte)) {
			/* store byte in buffer */
			uint8_t end = (uart_rxbuf.start + uart_rxbuf.count) & (UART_RX_BUFSIZE - 1);
			uart_rxbuf.data[end] = byte;
			uart_rxbuf.count++;
			
			if (!notified) {
				/* Notify that data has arrived */
				evt_enqueue(EVT_UART_RXCH);
				notified = true;
			}
			
			if (byte == '\n') {
				/* Notify that line is complete */
				evt_enqueue(EVT_UART_RXLN);
			}
		}
	} else {
		/* discard the byte :-( */
		RT_ASSERT(false);
	}
}

/* Initialization function */
void uart_init(void)
{
	/* set baud rate */
	uint16_t ubrr = (F_CPU / (16.0 * UART_BAUDRATE)) - 1;
	UBRRH = (uint8_t) (ubrr >> 8);
	UBRRL = (uint8_t) ubrr;
	
	/* enable transmitter and receiver */
	UCSRB = (1 << TXEN) | (1 << RXEN) | (1 << RXCIE);
	
	/* set frame format: 8 bits, no parity, 1 stop bit */
	UCSRC = (1 << URSEL) | (1 << UCSZ1) | (1 << UCSZ0);
	
	/* reset buffers */
	uart_rxbuf.count = 0;
	uart_rxbuf.start = 0;
	uart_txbuf.count = 0;
	uart_txbuf.start = 0;
	
	/* reset notified flag */
	notified = false;
}

/* Put one byte into output buffer (non-blocking) */
static error_t uart_tx_byte(const uint8_t byte)
{
	error_t result = ENOBUFS;

	/* Lock against UDRE interrupt */
	uint8_t sreg;
	irq_lock(&sreg);
	
	/* Check if there is space left */
	if (uart_txbuf.count < UART_TX_BUFSIZE) {
		/* Add the byte to the queue */
		uint8_t end = (uart_txbuf.start + uart_txbuf.count) & (UART_TX_BUFSIZE - 1);
		uart_txbuf.data[end] = byte;
		uart_txbuf.count++;
	
		/* Make sure the UDRE interrupt is enabled to start transmission */
		UCSRB |= (1 << UDRIE);
		result = ENOERR;
	}
	
	irq_unlock(sreg);
	return result;
}

/* Get one byte from input buffer (non-blocking) */
static error_t uart_rx_byte(uint8_t *buf)
{
	error_t result = EAGAIN;

	/* Lock against RXC interrupt */
	uint8_t sreg;
	irq_lock(&sreg);
	
	/* Check if there is data available */
	if (uart_rxbuf.count > 0) {
		/* Get one byte from the queue */
		*buf = uart_rxbuf.data[uart_rxbuf.start];
		uart_rxbuf.start = (uart_rxbuf.start + 1) & (UART_RX_BUFSIZE - 1);
		uart_rxbuf.count--;
		
		/* Reset notification flag */
		notified = false;
		result = ENOERR;
	}
	
	irq_unlock(sreg);
	return result;
}

/* Filter procedure for output */
inline bool uart_filter_output(uint8_t *byte)
{
	bool pass = true;
	
#ifdef UART_MODE_TEXT
	if (*byte == '\n') {
		/* Add CR to every LF */
		uart_putc('\r');
	}
#endif
	
#ifdef UART_MODE_BINARY
	/* Nothing to do */
#endif
	
	return pass;
}

/* Send one character (blocking) */
int uart_putc(int ch)
{
	bool error = false;
	bool done = false;
	uint8_t byte = ch;
	
	/* Check if character passes the filter */
	if (uart_filter_output(&byte)) {
		/* Attempt to place character in output buffer */
		while (!error && !done) {
			error_t result = uart_tx_byte(ch);
			switch (result) {
			case ENOERR:
				/* placed in buffer successfully */
				done = true;
				break;
			case ENOBUFS:
				/* retry */
				break;
			default:
				/* unknown error */
				error = true;
				break;
			}
		}
	}
	
	return error ? EOF : ch;
}

/* Send string (blocking) */
int uart_puts(const char *str)
{
	bool error = false;
	
	/* Send characters until end of string is reached or an error occurs */
	while (!error && (*str != '\0')) {
		if (uart_putc(*str) == EOF)
			error = true;
		str++;
	}
	
	/* ANSI states a non-negative result if successful,
	 * not necessarily the number of transmitted characters!
	 */
	return error ? EOF : 0;
}

/* Get one character (non-blocking) */
int uart_getc(void)
{
	uint8_t byte = 0;
	error_t result = uart_rx_byte(&byte);
	return (result == ENOERR) ? byte : EOF;
}

/* Get a whole line, but at most size-1 characters (non-blocking) */
char *uart_gets(char *buf, int size)
{
	int ch = EOF;
	bool success = false;
	char *result = buf;
	
	/* Loop until EOF is reached, line is complete, or buffer is full */
	do {
		/* Only add characters if there is space for the terminating '\0' */
		if (size > 1) {
			/* Get one character */
			ch = uart_getc();
			if (ch != EOF) {
				/* Append character to buffer */
				*(buf++) = ch;
				size--;
				/* Call is successful if at least one character was read */
				success = true;
			}
		}
	} while ((ch != EOF) && (ch != '\r') && (size > 1));
	
	/* If characters were read and size is non-zero, terminate string */
	if (success && (size > 0)) {
		/* Success, terminate string */
		*buf = '\0';
	} else {
		/* No success or invalid size argument */
		result = NULL;
	}
	
	return result;
}
