/*
 * Copyright (c) 2011, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

/*
 * uart.h -- UART services
 *
 * From the high-level view the UART behaves like a file stream opened in
 * either binary or text mode (depending on setting in config.h). In text mode
 * lines are terminated with LF ('\n'). See internal documentation for
 * specific details. One important difference is that the reading functions
 * never block but return EOF if no further data is available. To make
 * uart_gets work more pleasantly the EVT_UART_RXLN event was added.
 *
 * Public events:
 *   EVT_UART_RXCH  Data is available.
 *                  This event is sent only once until data has been read.
 *   EVT_UART_RXLN  A complete line is available (at least).
 *                  This event is sent for every complete line.
 *
 */

#ifndef UART_H
#define UART_H

#include <stddef.h>  // (to make sure NULL is defined)

/* Initialization function */
void uart_init(void);

/* Puts one character in the stream.
 * This function blocks if the output buffer is full.
 *
 * Parameters:
 *   ch    character to send
 *
 * Return value:
 *   The character on success, EOF on failure.
 */
int uart_putc(int ch);

/* Puts a string of characters in the stream.
 * This function blocks if the output buffer is full.
 * No line termination is added (e.g. like fputs, not like puts).
 *
 * Parameters:
 *   str    character string to send
 *
 * Return value:
 *   Non-negative on success, EOF on failure.
 */
int uart_puts(const char *str);

/* Gets one character from the stream.
 * This function never blocks, but returns EOF if no data is available.
 * Further calls to the function are allowed then (polling).
 *
 * Return value:
 *   Character read, or EOF if no data is available.
 */
int uart_getc(void);

/* Gets a string of characters from the stream. At most size-1 characters
 * are read. Reading stops after a LF character (which is placed in the
 * buffer, too). The resulting string is always zero-terminated, unless the
 * function fails (return value NULL).
 *
 * This function never blocks, but fails if no data is available.
 *
 * Parameters:
 *   buf     buffer to store string
 *   size    size of buffer in characters
 *
 * Return value:
 *   Pointer to the buffer on success, NULL on failure.
 */
char *uart_gets(char *buf, int size);

#endif
