/*
 * Copyright (c) 2011, Felix Opatz
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

/*
 * view.c -- optical and acoustic indication of heart beats
 *
 * This module provides the following indications of heart beats:
 *
 * LED   The LED is turned on with each detected heart beat. After a software
 *       timer has elapsed the LED is turned off.
 *
 * BEEP  The AC buzzer is turned on with each detected heart beat. It uses the
 *       same software timer as the LED to be turned off.
 *
 * UART  With each heart beat the current statistics are sent using the UART.
 *
 */

#include <stdio.h>
#include <avr/io.h>
#include <avr/pgmspace.h>

#include "config.h"
#include "event.h"
#include "pulse.h"
#include "uart.h"
#include "view.h"

/* Timeout counter, decremented each millisecond until zero */
static uint8_t view_timer;
/* Ticks since start for timestamp printout */
static uint32_t view_timestamp;

/* Callback function for 1ms timer events */
static void view_1ms_tick(evt_id_t id)
{
	/* Increment timestamp counter */
	view_timestamp++;

	/* Check if timeout hasn't elapsed */
	if (view_timer > 0) {
		/* Decrement timeout counter */
		view_timer --;
		/* If zero is reached the port pins are set to low */
		if (view_timer == 0) {
			LED_OREG &= ~(1 << LED_PIN);
			BEEP_OREG &= ~(1 << BEEP_PIN);
		}
	}
}

/* Print current statistics using the UART */
static void view_print_stats(void)
{
	PGM_P fmt_ticks = PSTR("%lu: ");
	PGM_P fmt_value_sep = PSTR("%u, ");
	PGM_P fmt_invalid_sep = PSTR("?, ");
	PGM_P fmt_value_end = PSTR("%u\r\n");
	PGM_P fmt_invalid_end = PSTR("?\r\n");
	
	char str[16];
	uint8_t value_u8;

	/* Tick counter */
	sprintf_P(str, fmt_ticks, view_timestamp);
	uart_puts(str);
	
	/* Current heart rate */
	if (pulse_get_curr(&value_u8))
		sprintf_P(str, fmt_value_sep, value_u8);
	else
		sprintf_P(str, fmt_invalid_sep);
	
	uart_puts(str);
	
	/* Average heart rate */
	if (pulse_get_avg(&value_u8))
		sprintf_P(str, fmt_value_end, value_u8);
	else
		sprintf_P(str, fmt_invalid_end);
	
	uart_puts(str);
}

/* Callback function for heart beat events */
static void view_on_beat(evt_id_t id)
{
	/* Set port pins to high */
	LED_OREG |= (1 << LED_PIN);
	BEEP_OREG |= (1 << BEEP_PIN);
	/* Start timeout by presetting the counter */
	view_timer = VIEW_ON_TIME;

	/* Print out statistics */
	view_print_stats();
}

/* Initialization function */
void view_init(void)
{
	/* Register for 1ms timer events */
	evt_register(view_1ms_tick, EVT_TMR_1MS);
	/* Register for heart beat events */
	evt_register(view_on_beat, EVT_PULSE_BEAT);
	
	/* Configure port pins as output and set pins to low */
	LED_DREG |= (1 << LED_PIN);
	LED_OREG &= ~(1 << LED_PIN);
	BEEP_DREG |= (1 << BEEP_PIN);
	BEEP_OREG &= ~(1 << BEEP_PIN);

	/* Reset timeout counter until first heart beat is detected */
	view_timer = 0;
	/* Reset tick counter */
	view_timestamp = 0;
}
